from uuid import uuid4

from django.db import models
from django.db.models import PROTECT
from edc_consent.managers import ConsentObjectsByCdefManager, CurrentSiteByCdefManager
from edc_consent.model_mixins.consent_version_model_mixin import (
    ConsentVersionModelMixin,
)
from edc_crf.model_mixins import CrfModelMixin, CrfWithActionModelMixin
from edc_identifier.managers import SubjectIdentifierManager
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierModelMixin
from edc_list_data.model_mixins import ListModelMixin
from edc_model.models import BaseUuidModel
from edc_registration.model_mixins import UpdatesOrCreatesRegistrationModelMixin
from edc_screening.model_mixins import ScreeningModelMixin
from edc_sites.model_mixins import SiteModelMixin
from edc_utils import get_utcnow
from edc_visit_tracking.model_mixins import SubjectVisitMissedModelMixin

from .consents import consent_v1


class DeathReport(BaseUuidModel):
    subject_identifier = models.CharField(max_length=50)


class OffStudy(BaseUuidModel):
    subject_identifier = models.CharField(max_length=50)

    offstudy_datetime = models.DateTimeField(default=get_utcnow)


class SubjectScreening(ScreeningModelMixin, BaseUuidModel):
    consent_definition = consent_v1
    objects = SubjectIdentifierManager()


class SubjectConsent(
    SiteModelMixin,
    ConsentVersionModelMixin,
    NonUniqueSubjectIdentifierModelMixin,
    UpdatesOrCreatesRegistrationModelMixin,
    BaseUuidModel,
):
    subject_identifier = models.CharField(max_length=50)

    consent_datetime = models.DateTimeField()

    dob = models.DateField(null=True)

    version = models.CharField(max_length=10)


class SubjectConsentV1(SubjectConsent):
    objects = ConsentObjectsByCdefManager()
    on_site = CurrentSiteByCdefManager()

    class Meta:
        proxy = True


class OnSchedule(SiteModelMixin, BaseUuidModel):
    subject_identifier = models.CharField(max_length=50)

    onschedule_datetime = models.DateTimeField(default=get_utcnow)


class OffSchedule(SiteModelMixin, BaseUuidModel):
    subject_identifier = models.CharField(max_length=50)

    offschedule_datetime = models.DateTimeField(default=get_utcnow)


class SubjectVisitMissedReasons(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "Subject Missed Visit Reasons"
        verbose_name_plural = "Subject Missed Visit Reasons"


class SubjectVisitMissed(
    SubjectVisitMissedModelMixin,
    CrfWithActionModelMixin,
    BaseUuidModel,
):
    missed_reasons = models.ManyToManyField(
        SubjectVisitMissedReasons, blank=True, related_name="+"
    )

    class Meta(
        SubjectVisitMissedModelMixin.Meta,
        BaseUuidModel.Meta,
    ):
        verbose_name = "Missed Visit Report"
        verbose_name_plural = "Missed Visit Report"


class Team(CrfModelMixin, BaseUuidModel):

    name = models.CharField(max_length=36, default=uuid4)

    class Meta(CrfModelMixin.Meta, BaseUuidModel.Meta):
        verbose_name = "Team"
        verbose_name_plural = "Teams"


class Venue(CrfModelMixin, BaseUuidModel):

    name = models.CharField(max_length=36, default=uuid4)

    class Meta(CrfModelMixin.Meta, BaseUuidModel.Meta):
        verbose_name = "Venue"
        verbose_name_plural = "Venues"


class Member(SiteModelMixin, BaseUuidModel):

    team = models.ForeignKey(Team, on_delete=PROTECT)

    player_name = models.CharField(max_length=36, default=uuid4)

    skill_level = models.CharField(max_length=36, default=uuid4)

    @property
    def subject_identifier(self):
        return self.team.subject_visit.subject_identifier

    @property
    def visit_code(self):
        return self.team.subject_visit.visit_code

    @property
    def visit_code_sequence(self):
        return self.team.subject_visit.visit_code_sequence

    class Meta(SiteModelMixin.Meta, BaseUuidModel.Meta):
        verbose_name = "Member"
        verbose_name_plural = "Members"


class Prn(CrfModelMixin, BaseUuidModel):

    name = models.CharField(max_length=36, default=uuid4)

    class Meta(CrfModelMixin.Meta, BaseUuidModel.Meta):
        verbose_name = "PRN"
        verbose_name_plural = "PRN"


class TeamWithDifferentFields(CrfModelMixin, BaseUuidModel):

    size = models.IntegerField(max_length=36)

    name = models.CharField(max_length=36, null=True, blank=False)

    color = models.CharField(max_length=36, null=True, blank=False)

    mood = models.CharField(max_length=36, default="good")

    class Meta(CrfModelMixin.Meta, BaseUuidModel.Meta):
        verbose_name = "TeamWithDifferentFields"
        verbose_name_plural = "TeamWithDifferentFields"



from __future__ import annotations

from .form_runner import FormRunner
from .site import site_form_runners

__all__ = ["get_form_runner"]


def get_form_runner(model_name: str | None = None, verbose: bool | None = None) -> FormRunner:
    if form_runner_cls := site_form_runners.registry.get(model_name):
        return form_runner_cls(model_name=model_name, verbose=verbose)
    return FormRunner(model_name=model_name, verbose=verbose)



__all__ = [
    "FormRunnerError",
    "FormRunnerModelAdminNotFound",
    "FormRunnerImproperlyConfigured",
    "FormRunnerRegisterError",
]


class FormRunnerError(Exception):
    pass


class FormRunnerModelAdminNotFound(Exception):
    pass


class FormRunnerModelFormNotFound(Exception):
    pass


class FormRunnerImproperlyConfigured(Exception):
    pass


class FormRunnerRegisterError(Exception):
    pass



from django import forms
from edc_form_validators import INVALID_ERROR, FormValidator, FormValidatorMixin

from .models import Member, Team, TeamWithDifferentFields


class MemberFormValidator(FormValidator):
    def clean(self) -> None:
        if self.cleaned_data.get("player_name") != "not-a-uuid":
            self.raise_validation_error({"player_name": "Cannot be a UUID"}, INVALID_ERROR)


class TeamFormValidator(FormValidator):
    def clean(self) -> None:
        if self.cleaned_data.get("name") != "not-a-uuid":
            self.raise_validation_error({"name": "Cannot be a UUID"}, INVALID_ERROR)


class MemberForm(FormValidatorMixin, forms.ModelForm):
    form_validator_cls = MemberFormValidator

    class Meta:
        model = Member
        fields = "__all__"


class TeamForm(FormValidatorMixin, forms.ModelForm):
    form_validator_cls = TeamFormValidator

    class Meta:
        model = Team
        fields = "__all__"


class TeamWithDifferentFieldsForm(FormValidatorMixin, forms.ModelForm):
    form_validator_cls = None

    def clean(self):
        if self.cleaned_data.get("name") != "not-a-uuid":
            raise forms.ValidationError({"name": "Cannot be a UUID"}, INVALID_ERROR)

    class Meta:
        model = TeamWithDifferentFields
        fields = "__all__"



from __future__ import annotations

import html
import uuid
from typing import TYPE_CHECKING, Any

from bs4 import BeautifulSoup
from django.apps import apps as django_apps
from django.db.models import ForeignKey, ManyToManyField, Model, OneToOneField, QuerySet
from django.forms import ModelForm
from edc_utils import get_utcnow
from tqdm import tqdm

from .exceptions import FormRunnerModelAdminNotFound, FormRunnerModelFormNotFound
from .utils import get_modeladmin_cls

if TYPE_CHECKING:
    from .models import Issue

__all__ = ["FormRunner"]


class FormRunner:
    """Rerun modelform validation on all instances of a model"""

    model_name: str | None = None
    issue_model = "edc_form_runners.issue"
    extra_formfields: list[str] | None = None
    exclude_formfields: list[str] | None = None

    def __init__(
        self,
        model_name: str | None = None,
        src_filter_options: dict[str, Any] | None = None,
        verbose: bool | None = None,
    ) -> None:
        self.messages = {}
        self.session_id = uuid.uuid4()
        self.session_datetime = get_utcnow()
        self.verbose = verbose
        self.model_name = self.model_name or model_name
        self.modeladmin_cls = get_modeladmin_cls(self.model_name)
        if not self.modeladmin_cls:
            raise FormRunnerModelAdminNotFound(
                "No modeladmin class found. Is this model registered with Admin? "
                f"Got `{model_name}`."
            )
        if self.modeladmin_cls.form == ModelForm:
            raise FormRunnerModelFormNotFound(
                "ModelAdmin does not have a custom form. Nothing to do. "
                f"Got `{model_name}`."
            )
        self.src_model_cls = self.modeladmin_cls.model

        # note: modeladmin_cls must declare a custom ModelForm
        self.modelform_cls = self.modeladmin_cls.form

        self.src_filter_options = src_filter_options

    def __repr__(self):
        return f"{self.__class__}({self.model_name})"

    def __str__(self):
        return self.model_name

    def run_all(self) -> None:
        total = self.src_qs.count()
        for src_obj in tqdm(self.src_qs, total=total):
            self.issue_model_cls.objects.filter(**self.unique_opts(src_obj)).delete()
            self.run_one(src_obj=src_obj, skip_delete=True)
        for k, v in self.messages.items():
            print(f"Warning: {k}: {v}")

    def run_one(self, src_obj: Model, skip_delete: bool | None = None) -> None:
        if not skip_delete:
            self.issue_model_cls.objects.filter(**self.unique_opts(src_obj)).delete()
        data = self.get_form_data(src_obj)
        form = self.modelform_cls(data, instance=src_obj)
        form.is_valid()
        errors = {
            k: v for k, v in form._errors.items() if k not in self.get_exclude_formfields()
        }
        if errors:
            for fldname, errmsg in errors.items():
                if fldname in self.fieldset_fields:
                    issue_obj = self.write_to_db(fldname, errmsg, src_obj)
                    if self.verbose:
                        self.print(str(issue_obj))

    @property
    def issue_model_cls(self) -> Issue:
        return django_apps.get_model(self.issue_model)

    @property
    def fieldset_fields(self) -> list[str]:
        fields = []
        if self.modeladmin_cls.form != ModelForm:
            if getattr(self.modeladmin_cls, "fieldsets", None):
                for fieldset in self.modeladmin_cls.fieldsets:
                    _, data = fieldset
                    fields.extend(data.get("fields"))
            else:
                self.messages.update(
                    {
                        self.model_name: (
                            "ModelAdmin fieldsets not defined. Using ModelForm fields."
                        )
                    }
                )

                fields = [
                    k for k, v in getattr(self.modeladmin_cls.form(), "fields", {}).items()
                ]
        fields = list(set(fields))
        return fields

    def write_to_db(self, fldname: str, errmsg: Any, src_obj: Any) -> Issue:
        raw_message = html.unescape(errmsg.as_text())
        message = BeautifulSoup(raw_message, "html.parser").text
        try:
            response = getattr(src_obj, fldname)
        except AttributeError:
            response = None
        return self.issue_model_cls.objects.create(
            session_id=self.session_id,
            session_datetime=self.session_datetime,
            raw_message=raw_message,
            message=message,
            short_message=message[:250],
            response=str(response),
            src_id=src_obj.id,
            src_revision=src_obj.revision,
            src_report_datetime=getattr(src_obj, "report_datetime", None),
            src_created_datetime=src_obj.created,
            src_modified_datetime=src_obj.modified,
            src_user_created=src_obj.user_created,
            src_user_modified=src_obj.user_modified,
            field_name=fldname,
            site=src_obj.site,
            extra_formfields=",".join(self.get_extra_formfields()),
            exclude_formfields=",".join(self.get_exclude_formfields()),
            **self.unique_opts(src_obj),
        )

    def unique_opts(self, src_obj: Model) -> dict[str, Any]:
        """Note: unique constraint includes `field_name`"""
        model_obj_or_related_visit = src_obj
        get_related_visit_model_attr = getattr(src_obj, "related_visit_model_attr", None)
        if (
            get_related_visit_model_attr
            and get_related_visit_model_attr()
            and src_obj.related_visit
        ):
            model_obj_or_related_visit = src_obj.related_visit
        subject_identifier = model_obj_or_related_visit.subject_identifier
        opts = dict(
            label_lower=src_obj._meta.label_lower,
            panel_name=self.get_panel_name(src_obj),
            verbose_name=src_obj._meta.verbose_name,
            subject_identifier=subject_identifier,
        )
        for fldname in [
            "visit_code",
            "visit_code_sequence",
            "visit_schedule_name",
            "schedule_name",
        ]:
            value = getattr(model_obj_or_related_visit, fldname, None)
            if value is not None:
                opts.update({fldname: value})
        return opts

    @staticmethod
    def get_panel_name(src_obj) -> str | None:
        panel = getattr(src_obj, "panel", None)
        return getattr(panel, "name", None)

    @property
    def src_qs(self) -> QuerySet:
        return self.src_model_cls.objects.filter(**(self.get_src_filter_options() or {}))

    def get_form_data(self, src_obj: Any) -> dict[str, Any]:
        data = {
            k: v
            for k, v in src_obj.__dict__.items()
            if not k.startswith("_") and not k.endswith("_id")
        }
        for fld_cls in src_obj._meta.get_fields():
            if isinstance(fld_cls, (ForeignKey, OneToOneField)):
                try:
                    obj_fld_id = getattr(src_obj, fld_cls.name).id
                except AttributeError:
                    rel_obj = None
                else:
                    rel_obj = fld_cls.related_model.objects.get(id=obj_fld_id)
                data.update({fld_cls.name: rel_obj})
            elif isinstance(fld_cls, (ManyToManyField,)):
                data.update({fld_cls.name: getattr(src_obj, fld_cls.name).all()})
            else:
                pass
        try:
            data.update(subject_visit=src_obj.subject_visit)
        except AttributeError:
            data.update(subject_identifier=src_obj.subject_identifier)
        for extra_formfield in self.get_extra_formfields():
            data.update({extra_formfield: getattr(src_obj, extra_formfield)})
        return data

    def get_src_filter_options(self) -> dict[str, Any]:
        return self.src_filter_options

    def get_extra_formfields(self) -> list[str]:
        return self.extra_formfields or []

    def get_exclude_formfields(self) -> list[str]:
        return self.exclude_formfields or []

    def print(self, msg: str) -> None:
        if self.verbose:
            print(msg)



# Generated by Django 4.2.7 on 2023-12-05 18:21

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_form_runners", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="issue",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Issue",
                "verbose_name_plural": "Issues",
            },
        ),
        migrations.RemoveConstraint(
            model_name="issue",
            name="unique_label_lower_subject_identifier_etc",
        ),
        migrations.RemoveIndex(
            model_name="issue",
            name="edc_form_ru_label_l_ce3dda_idx",
        ),
        migrations.RemoveIndex(
            model_name="issue",
            name="edc_form_ru_subject_10bb16_idx",
        ),
        migrations.RenameField(
            model_name="issue",
            old_name="ignore_formfields",
            new_name="exclude_formfields",
        ),
        migrations.RemoveField(
            model_name="issue",
            name="status",
        ),
        migrations.AddField(
            model_name="issue",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="issue",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="issue",
            name="panel_name",
            field=models.CharField(max_length=50, null=True),
        ),
        migrations.AddField(
            model_name="issue",
            name="schedule_name",
            field=models.CharField(max_length=25, null=True),
        ),
        migrations.AddField(
            model_name="issue",
            name="src_created_datetime",
            field=models.DateTimeField(null=True),
        ),
        migrations.AddField(
            model_name="issue",
            name="src_user_created",
            field=models.CharField(max_length=150, null=True),
        ),
        migrations.AddField(
            model_name="issue",
            name="visit_schedule_name",
            field=models.CharField(max_length=25, null=True),
        ),
        migrations.AlterField(
            model_name="issue",
            name="label_lower",
            field=models.CharField(max_length=50),
        ),
        migrations.AlterField(
            model_name="issue",
            name="subject_identifier",
            field=models.CharField(max_length=50),
        ),
        migrations.AlterField(
            model_name="issue",
            name="visit_code",
            field=models.CharField(max_length=25),
        ),
        migrations.AlterField(
            model_name="issue",
            name="visit_code_sequence",
            field=models.IntegerField(),
        ),
        migrations.AddIndex(
            model_name="issue",
            index=models.Index(
                fields=["modified", "created"], name="edc_form_ru_modifie_486e31_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="issue",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_form_ru_user_mo_fd5ddf_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="issue",
            index=models.Index(
                fields=["label_lower", "field_name", "panel_name", "short_message"],
                name="edc_form_ru_label_l_c7917b_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="issue",
            index=models.Index(
                fields=[
                    "subject_identifier",
                    "visit_code",
                    "visit_code_sequence",
                    "label_lower",
                    "panel_name",
                    "short_message",
                ],
                name="edc_form_ru_subject_6e8602_idx",
            ),
        ),
        migrations.AddConstraint(
            model_name="issue",
            constraint=models.UniqueConstraint(
                fields=(
                    "subject_identifier",
                    "label_lower",
                    "panel_name",
                    "visit_code",
                    "visit_code_sequence",
                    "visit_schedule_name",
                    "schedule_name",
                    "field_name",
                ),
                name="unique_label_lower_subject_identifier_etc",
            ),
        ),
    ]



from django.db.models.signals import post_save
from django.dispatch import receiver
from edc_crf.model_mixins import CrfModelMixin, CrfNoManagerModelMixin
from edc_lab.model_mixins import RequisitionModelMixin
from edc_model.models import BaseUuidModel

from ..exceptions import FormRunnerModelAdminNotFound, FormRunnerModelFormNotFound
from ..get_form_runner_by_src_id import get_form_runner_by_src_id
from ..utils import get_edc_form_runners_enabled


@receiver(
    post_save,
    weak=False,
    dispatch_uid="update_issue_on_post_save",
)
def update_issue_on_post_save(sender, instance, raw, created, update_fields, **kwargs):
    """Updates Issue data on post-save"""
    if not raw and not update_fields:
        if get_edc_form_runners_enabled():
            if isinstance(
                instance, (CrfModelMixin, CrfNoManagerModelMixin, RequisitionModelMixin)
            ) and isinstance(instance, (BaseUuidModel,)):
                try:
                    runner = get_form_runner_by_src_id(
                        src_id=instance.id,
                        model_name=sender._meta.label_lower,
                    )
                except (FormRunnerModelAdminNotFound, FormRunnerModelFormNotFound):
                    pass
                else:
                    runner.run_one()



from django.contrib import admin
from edc_model_admin.admin_site import EdcAdminSite
from edc_model_admin.dashboard import (
    ModelAdminCrfDashboardMixin,
    ModelAdminDashboardMixin,
)
from edc_model_admin.history import SimpleHistoryAdmin
from edc_sites.admin import SiteModelAdminMixin

from .forms import MemberForm, TeamForm, TeamWithDifferentFieldsForm
from .models import Member, Team, TeamWithDifferentFields, Venue

my_admin_site = EdcAdminSite(name="my_admin_site")


class MemberInlineAdmin(admin.TabularInline):
    model = Member
    form = MemberForm

    extra = 0


@admin.register(Member, site=my_admin_site)
class MembersAdmin(SiteModelAdminMixin, ModelAdminDashboardMixin, SimpleHistoryAdmin):
    form = MemberForm


@admin.register(Team, site=my_admin_site)
class TeamAdmin(SiteModelAdminMixin, ModelAdminCrfDashboardMixin, SimpleHistoryAdmin):
    form = TeamForm

    inlines = [MemberInlineAdmin]

    fieldsets = ((None, ({"fields": ("name", "created", "modified")})),)


@admin.register(TeamWithDifferentFields, site=my_admin_site)
class TeamWithDifferentFieldsAdmin(
    SiteModelAdminMixin, ModelAdminCrfDashboardMixin, SimpleHistoryAdmin
):
    form = TeamWithDifferentFieldsForm

    # do not include "name" to show that the field is ignored
    # by FormRunner.run even though blank=False
    fieldsets = ((None, ({"fields": ("size", "color", "mood")})),)


@admin.register(Venue, site=my_admin_site)
class VenueAdmin(SiteModelAdminMixin, ModelAdminDashboardMixin, SimpleHistoryAdmin):
    pass



from datetime import datetime
from uuid import uuid4
from zoneinfo import ZoneInfo

import time_machine
from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase, override_settings
from edc_appointment.models import Appointment
from edc_appointment.tests.helper import Helper
from edc_consent.site_consents import site_consents
from edc_facility import import_holidays
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED
from edc_visit_tracking.models import SubjectVisit

from edc_form_runners.exceptions import FormRunnerModelFormNotFound
from edc_form_runners.form_runner import FormRunner
from edc_form_runners.models import Issue
from form_runners_app.consents import consent_v1
from form_runners_app.models import Member, Team, TeamWithDifferentFields, Venue
from form_runners_app.visit_schedules import visit_schedule


@override_settings(SITE_ID=10)
@time_machine.travel(datetime(2019, 6, 11, 8, 00, tzinfo=ZoneInfo("UTC")))
class TestRunners(TestCase):

    helper_cls = Helper

    @classmethod
    def setUpTestData(cls):
        import_holidays()
        site_consents.registry = {}
        site_consents.register(consent_v1)

    def setUp(self):
        site_consents.registry = {}
        site_consents.register(consent_v1)
        site_visit_schedules._registry = {}

        site_visit_schedules.loaded = False
        site_visit_schedules.register(visit_schedule)

        self.subject_identifier = "1235"
        self.helper = self.helper_cls(subject_identifier=self.subject_identifier)
        self.visit_schedule_name = "visit_schedule"
        self.schedule_name = "schedule"

        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule", schedule_name="schedule"
        )

    def test_appointment(self):
        form_runner = FormRunner(model_name="edc_appointment.appointment")
        form_runner.run_all()

    def test_crfs(self):
        for appointment in Appointment.objects.all().order_by("timepoint_datetime"):
            subject_visit = SubjectVisit.objects.create(
                appointment=appointment,
                subject_identifier=self.subject_identifier,
                reason=SCHEDULED,
            )
            TeamWithDifferentFields.objects.create(subject_visit=subject_visit, size=11)
            Venue.objects.create(subject_visit=subject_visit, name=uuid4())
            team = Team.objects.create(subject_visit=subject_visit, name=uuid4())
            Member.objects.create(team=team)
            Member.objects.create(team=team)
            Member.objects.create(team=team)

        # raise on VenueModelAdmin has no custom ModelForm
        self.assertRaises(
            FormRunnerModelFormNotFound, FormRunner, model_name="form_runners_app.venue"
        )

        # run to find name field may not be a UUID
        # see form validator
        form_runner = FormRunner(model_name="form_runners_app.team")
        form_runner.run_all()
        self.assertEqual(Issue.objects.all().count(), 1)
        try:
            Issue.objects.get(
                label_lower="form_runners_app.team",
                visit_code="1000",
                visit_code_sequence=0,
                field_name="name",
                message__icontains="Cannot be a UUID",
            )
        except ObjectDoesNotExist:
            self.fail("Issue model instance unexpectedly does not exist")

        # run to find player_name field may not be a UUID
        # see form validator
        form_runner = FormRunner(model_name="form_runners_app.member")
        form_runner.run_all()
        try:
            Issue.objects.get(
                label_lower="form_runners_app.member",
                visit_code="1000",
                visit_code_sequence=0,
                field_name="player_name",
                message__icontains="Cannot be a UUID",
            )
        except ObjectDoesNotExist:
            self.fail("Issue model instance unexpectedly does not exist")

        # run to assert ignores `name` field because it IS NOT IN admin fieldsets
        # even though the model instance field class has blank=False.
        form_runner = FormRunner(model_name="form_runners_app.teamwithdifferentfields")
        form_runner.run_all()
        try:
            Issue.objects.get(
                label_lower="form_runners_app.teamwithdifferentfields",
                field_name="name",
            )
        except ObjectDoesNotExist:
            pass
        else:
            self.fail("Issue model instance unexpectedly does not exist")
        # assert does not ignore `color` field because it IS IN admin fieldsets
        # and the model instance field class has blank=False.
        try:
            Issue.objects.get(
                label_lower="form_runners_app.teamwithdifferentfields",
                field_name="color",
            )
        except ObjectDoesNotExist:
            self.fail("Issue model instance unexpectedly does not exist")



import sys
from datetime import datetime
from pathlib import Path
from zoneinfo import ZoneInfo

from dateutil.relativedelta import relativedelta
from edc_test_settings.default_test_settings import DefaultTestSettings

app_name = "edc_form_runners"
base_dir = Path(__file__).absolute().parent.parent.parent

project_settings = DefaultTestSettings(
    calling_file=__file__,
    BASE_DIR=base_dir,
    APP_NAME=app_name,
    ETC_DIR=base_dir / app_name / "tests" / "etc",
    SILENCED_SYSTEM_CHECKS=[
        "sites.E101",
        "edc_navbar.E002",
        "edc_navbar.E003",
        "edc_sites.E001",
    ],
    EDC_NAVBAR_DEFAULT="form_runners_app",
    EDC_AUTH_CODENAMES_WARN_ONLY=True,
    EDC_AUTH_SKIP_SITE_AUTHS=True,
    EDC_AUTH_SKIP_AUTH_UPDATER=True,
    SUBJECT_VISIT_MODEL="edc_visit_tracking.subjectvisit",
    SUBJECT_VISIT_MISSED_MODEL="edc_visit_tracking.subjectvisitmissed",
    SUBJECT_SCREENING_MODEL="form_runners_app.subjectscreening",
    EDC_SITES_REGISTER_DEFAULT=False,
    ROOT_URLCONF="form_runners_app.urls",
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=(
        datetime(2018, 5, 15, 0, 0, 0, tzinfo=ZoneInfo("UTC")) - relativedelta(years=1)
    ),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=(
        datetime(2022, 5, 14, 0, 0, 0, tzinfo=ZoneInfo("UTC")) - relativedelta(years=1)
    ),
    INSTALLED_APPS=[
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.staticfiles",
        "django.contrib.sites",
        "django_crypto_fields.apps.AppConfig",
        "django_revision.apps.AppConfig",
        "edc_appointment.apps.AppConfig",
        "edc_auth.apps.AppConfig",
        "edc_action_item.apps.AppConfig",
        "edc_adverse_event.apps.AppConfig",
        "adverse_event_app.apps.AppConfig",
        "edc_consent.apps.AppConfig",
        "edc_data_manager.apps.AppConfig",
        "edc_dashboard.apps.AppConfig",
        "edc_device.apps.AppConfig",
        "edc_identifier.apps.AppConfig",
        "edc_facility.apps.AppConfig",
        "edc_lab.apps.AppConfig",
        "edc_metadata.apps.AppConfig",
        "edc_notification.apps.AppConfig",
        "edc_registration.apps.AppConfig",
        "edc_review_dashboard.apps.AppConfig",
        "edc_sites.apps.AppConfig",
        "edc_subject_dashboard.apps.AppConfig",
        "edc_visit_schedule.apps.AppConfig",
        "edc_visit_tracking.apps.AppConfig",
        "edc_timepoint.apps.AppConfig",
        "edc_form_runners.apps.AppConfig",
        "form_runners_app.apps.AppConfig",
        "edc_appconfig.apps.AppConfig",
    ],
    add_dashboard_middleware=True,
    use_test_urls=True,
).settings

for k, v in project_settings.items():
    setattr(sys.modules[__name__], k, v)



from __future__ import annotations

from typing import TYPE_CHECKING

from django import template

from edc_form_runners.utils import get_form_runner_issues

if TYPE_CHECKING:
    from django.db import models
    from edc_crf.model_mixins import CrfModelMixin
    from edc_metadata.models import CrfMetadata, RequisitionMetadata

    class Model(CrfModelMixin, models.Model): ...


register = template.Library()


@register.inclusion_tag("edc_form_runners/form_runner_issues.html")
def show_form_runner_issues(metadata_model_obj: CrfMetadata | RequisitionMetadata):
    messages = []
    if metadata_model_obj and metadata_model_obj.model_instance:
        model_obj = metadata_model_obj.model_instance
        related_visit = getattr(model_obj, model_obj.related_visit_model_attr())
        panel_name = getattr(metadata_model_obj, "panel_name", None)
        qs = get_form_runner_issues(
            model_obj._meta.label_lower, related_visit, panel_name=panel_name
        )
        messages = [f"{issue.message} [{issue.field_name}]" for issue in qs]
    return dict(form_runner_issues="<BR>".join(messages))
